@@ -8,8 +8,9 @@
 //! Various storage and representation related types
 
 use crate::node_state::NodeLabel;
+use crate::storage::Storable;
 use serde::{Deserialize, Serialize};
-use winter_crypto::Hasher;
+use std::convert::TryInto;
 
 /// Various elements that can be stored
 #[derive(PartialEq, Eq, Debug, Hash, Clone, Copy)]
@@ -41,11 +42,16 @@ pub struct ValueStateKey(pub String, pub u64);
 #[derive(Debug, Serialize, Deserialize, Clone, PartialEq, Eq, Hash)]
 #[serde(bound = "")]
 pub struct ValueState {
-    pub(crate) plaintext_val: Values, // This needs to be the plaintext value, to discuss
-    pub(crate) version: u64,          // to discuss
-    pub(crate) label: NodeLabel,
-    pub(crate) epoch: u64,
-    pub(crate) username: AkdKey,
+    /// The plaintext value of the user information in the directory
+    pub plaintext_val: Values, // This needs to be the plaintext value, to discuss
+    /// The version of the user's value-state
+    pub version: u64, // to discuss
+    /// The Node Label
+    pub label: NodeLabel,
+    /// The epoch this value state was published in
+    pub epoch: u64,
+    /// The username associated to this value state (username + epoch is the record key)
+    pub username: AkdKey,
 }
 
 impl crate::storage::Storable for ValueState {
@@ -58,6 +64,33 @@ impl crate::storage::Storable for ValueState {
     fn get_id(&self) -> ValueStateKey {
         ValueStateKey(self.username.0.clone(), self.epoch)
     }
+
+    fn get_full_binary_key_id(key: &ValueStateKey) -> Vec<u8> {
+        let mut result = vec![StorageType::ValueState as u8];
+        let epoch_bytes = key.1.to_be_bytes();
+        for byte in &epoch_bytes {
+            result.push(*byte);
+        }
+        let uname_bytes = key.0.as_bytes();
+        for byte in uname_bytes {
+            result.push(*byte);
+        }
+
+        result
+    }
+
+    fn key_from_full_binary(bin: &[u8]) -> Result<ValueStateKey, String> {
+        if bin.len() < 10 {
+            return Err("Not enough bytes to form a proper key".to_string());
+        }
+        let epoch_bytes: [u8; 8] = bin[1..=8].try_into().expect("Slice with incorrect length");
+        let epoch = u64::from_be_bytes(epoch_bytes);
+        if let Ok(username) = std::str::from_utf8(&bin[9..]) {
+            Ok(ValueStateKey(username.to_string(), epoch))
+        } else {
+            Err("Invalid string format".to_string())
+        }
+    }
 }
 
 impl ValueState {
@@ -78,20 +111,16 @@ impl ValueState {
     }
 }
 
-impl evmap::ShallowCopy for ValueState {
-    unsafe fn shallow_copy(&self) -> std::mem::ManuallyDrop<Self> {
-        std::mem::ManuallyDrop::new(self.clone())
-    }
-}
-
 /// Data associated with a given key. That is all the states at the various epochs
 #[derive(Debug, Serialize, Deserialize, Clone, PartialEq, Eq, Hash)]
 #[serde(bound = "")]
 pub struct KeyData {
-    pub(crate) states: Vec<ValueState>,
+    /// The vector of states of key data for a given AkdKey
+    pub states: Vec<ValueState>,
 }
 
 /// Used to retrieve a value's state, for a given key
+#[derive(std::fmt::Debug, Clone, Copy)]
 pub enum ValueStateRetrievalFlag {
     /// Specific version
     SpecificVersion(u64),
@@ -101,25 +130,44 @@ pub enum ValueStateRetrievalFlag {
     LeqEpoch(u64),
     /// State at the latest epoch
     MaxEpoch,
-    /// State at the latest version
-    MaxVersion,
     /// State at the earliest epoch
     MinEpoch,
-    /// Earliest version
-    MinVersion,
 }
 
 // == New Data Retrieval Logic == //
 
 /// This needs to be PUBLIC public, since anyone implementing a data-layer will need
 /// to be able to access this and all the internal types
-pub enum DbRecord<H: Hasher + Sync + Send> {
+#[derive(Serialize, Deserialize, Debug, PartialEq, Eq)]
+pub enum DbRecord {
     /// An Azks
-    Azks(crate::append_only_zks::Azks<H>),
+    Azks(crate::append_only_zks::Azks),
     /// A HistoryTreeNode
-    HistoryTreeNode(crate::history_tree_node::HistoryTreeNode<H>),
+    HistoryTreeNode(crate::history_tree_node::HistoryTreeNode),
     /// A HistoryNodeState
-    HistoryNodeState(crate::node_state::HistoryNodeState<H>),
+    HistoryNodeState(crate::node_state::HistoryNodeState),
     /// The state of the value for a particular key.
     ValueState(crate::storage::types::ValueState),
 }
+
+impl Clone for DbRecord {
+    fn clone(&self) -> Self {
+        match &self {
+            DbRecord::Azks(azks) => DbRecord::Azks(azks.clone()),
+            DbRecord::HistoryNodeState(state) => DbRecord::HistoryNodeState(state.clone()),
+            DbRecord::HistoryTreeNode(node) => DbRecord::HistoryTreeNode(node.clone()),
+            DbRecord::ValueState(state) => DbRecord::ValueState(state.clone()),
+        }
+    }
+}
+
+impl DbRecord {
+    pub(crate) fn get_full_binary_id(&self) -> Vec<u8> {
+        match &self {
+            DbRecord::Azks(azks) => azks.get_full_binary_id(),
+            DbRecord::HistoryNodeState(state) => state.get_full_binary_id(),
+            DbRecord::HistoryTreeNode(node) => node.get_full_binary_id(),
+            DbRecord::ValueState(state) => state.get_full_binary_id(),
+        }
+    }
+}
