@@ -35,7 +35,7 @@ async fn test_set_child_without_hash_at_root() -> Result<(), HistoryTreeNodeErro
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(ep)).await?;
     let child_hist_node_1 =
-        HistoryChildState::new::<Blake3>(1, NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
     root.write_to_storage(&db).await?;
     root.set_child::<_, Blake3>(&db, ep, &(Direction::Some(1), child_hist_node_1.clone()))
         .await?;
@@ -53,7 +53,10 @@ async fn test_set_child_without_hash_at_root() -> Result<(), HistoryTreeNodeErro
         root.get_latest_epoch().unwrap_or(0) == 1,
         "Latest epochs don't match!"
     );
-    assert!(root.epochs.len() == 1, "Ask yourself some pressing questions, such as: Why are there random extra epochs in the root?");
+    assert!(
+        root.birth_epoch == root.last_epoch,
+        "How would the last epoch be different from the birth epoch without an update?"
+    );
 
     Ok(())
 }
@@ -64,9 +67,9 @@ async fn test_set_children_without_hash_at_root() -> Result<(), HistoryTreeNodeE
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(ep)).await?;
     let child_hist_node_1 =
-        HistoryChildState::new::<Blake3>(1, NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
     let child_hist_node_2: HistoryChildState =
-        HistoryChildState::new::<Blake3>(2, NodeLabel::new(0, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(0, 1), Blake3::hash(&[0u8]), ep);
     root.write_to_storage(&db).await?;
     assert!(
         root.set_child::<_, Blake3>(&db, ep, &(Direction::Some(1), child_hist_node_1.clone()),)
@@ -103,7 +106,10 @@ async fn test_set_children_without_hash_at_root() -> Result<(), HistoryTreeNodeE
     }
     let latest_ep = root.get_latest_epoch();
     assert!(latest_ep.unwrap_or(0) == 1, "Latest epochs don't match!");
-    assert!(root.epochs.len() == 1, "Ask yourself some pressing questions, such as: Why are there random extra epochs in the root?");
+    assert!(
+        root.birth_epoch == root.last_epoch,
+        "How would the last epoch be different from the birth epoch without an update?"
+    );
 
     Ok(())
 }
@@ -114,9 +120,9 @@ async fn test_set_children_without_hash_multiple_at_root() -> Result<(), History
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(ep)).await?;
     let child_hist_node_1 =
-        HistoryChildState::new::<Blake3>(1, NodeLabel::new(11, 2), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(11, 2), Blake3::hash(&[0u8]), ep);
     let child_hist_node_2: HistoryChildState =
-        HistoryChildState::new::<Blake3>(2, NodeLabel::new(00, 2), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(00, 2), Blake3::hash(&[0u8]), ep);
     root.write_to_storage(&db).await?;
     assert!(
         root.set_child::<_, Blake3>(&db, ep, &(Direction::Some(1), child_hist_node_1))
@@ -134,9 +140,9 @@ async fn test_set_children_without_hash_multiple_at_root() -> Result<(), History
     ep = 2;
 
     let child_hist_node_3: HistoryChildState =
-        HistoryChildState::new::<Blake3>(1, NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
     let child_hist_node_4: HistoryChildState =
-        HistoryChildState::new::<Blake3>(2, NodeLabel::new(0, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(0, 1), Blake3::hash(&[0u8]), ep);
     root.write_to_storage(&db).await?;
     assert!(
         root.set_child::<_, Blake3>(&db, ep, &(Direction::Some(1), child_hist_node_3.clone()),)
@@ -173,7 +179,10 @@ async fn test_set_children_without_hash_multiple_at_root() -> Result<(), History
     }
     let latest_ep = root.get_latest_epoch();
     assert!(latest_ep.unwrap_or(0) == 2, "Latest epochs don't match!");
-    assert!(root.epochs.len() == 2, "Ask yourself some pressing questions, such as: Why are there random extra epochs in the root?");
+    assert!(
+        root.birth_epoch < root.last_epoch,
+        "How is the last epoch not higher than the birth epoch after an udpate?"
+    );
 
     Ok(())
 }
@@ -184,9 +193,9 @@ async fn test_get_child_at_existing_epoch_multiple_at_root() -> Result<(), Histo
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(ep)).await?;
     let child_hist_node_1 =
-        HistoryChildState::new::<Blake3>(1, NodeLabel::new(11, 2), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(11, 2), Blake3::hash(&[0u8]), ep);
     let child_hist_node_2: HistoryChildState =
-        HistoryChildState::new::<Blake3>(2, NodeLabel::new(00, 2), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(00, 2), Blake3::hash(&[0u8]), ep);
     root.write_to_storage(&db).await?;
     assert!(
         root.set_child::<_, Blake3>(&db, ep, &(Direction::Some(1), child_hist_node_1.clone()),)
@@ -204,9 +213,9 @@ async fn test_get_child_at_existing_epoch_multiple_at_root() -> Result<(), Histo
     ep = 2;
 
     let child_hist_node_3: HistoryChildState =
-        HistoryChildState::new::<Blake3>(1, NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(1, 1), Blake3::hash(&[0u8]), ep);
     let child_hist_node_4: HistoryChildState =
-        HistoryChildState::new::<Blake3>(2, NodeLabel::new(0, 1), Blake3::hash(&[0u8]), ep);
+        HistoryChildState::new::<Blake3>(NodeLabel::new(0, 1), Blake3::hash(&[0u8]), ep);
     assert!(
         root.set_child::<_, Blake3>(&db, ep, &(Direction::Some(1), child_hist_node_3.clone()),)
             .await
@@ -242,7 +251,10 @@ async fn test_get_child_at_existing_epoch_multiple_at_root() -> Result<(), Histo
     }
     let latest_ep = root.get_latest_epoch();
     assert!(latest_ep.unwrap_or(0) == 2, "Latest epochs don't match!");
-    assert!(root.epochs.len() == 2, "Ask yourself some pressing questions, such as: Why are there random extra epochs in the root?");
+    assert!(
+        root.birth_epoch < root.last_epoch,
+        "How is the last epoch not higher than the birth epoch after an udpate?"
+    );
 
     Ok(())
 }
@@ -254,15 +266,13 @@ pub async fn test_get_child_at_epoch_at_root() -> Result<(), HistoryTreeNodeErro
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(init_ep)).await?;
 
-    for ep in 0u64..3u64 {
+    for ep in 0..3 {
         let child_hist_node_1 = HistoryChildState::new::<Blake3>(
-            ep.try_into().unwrap(),
             NodeLabel::new(0b1u64 << ep.clone(), ep.try_into().unwrap()),
             Blake3::hash(&[0u8]),
             2 * ep,
         );
         let child_hist_node_2: HistoryChildState = HistoryChildState::new::<Blake3>(
-            ep.try_into().unwrap(),
             NodeLabel::new(0, ep.clone().try_into().unwrap()),
             Blake3::hash(&[0u8]),
             2 * ep,
@@ -277,7 +287,6 @@ pub async fn test_get_child_at_epoch_at_root() -> Result<(), HistoryTreeNodeErro
     let ep_existing = 0u64;
 
     let child_hist_node_1 = HistoryChildState::new::<Blake3>(
-        0,
         NodeLabel::new(
             0b1u64 << ep_existing.clone(),
             ep_existing.try_into().unwrap(),
@@ -286,7 +295,6 @@ pub async fn test_get_child_at_epoch_at_root() -> Result<(), HistoryTreeNodeErro
         2 * ep_existing,
     );
     let child_hist_node_2: HistoryChildState = HistoryChildState::new::<Blake3>(
-        0,
         NodeLabel::new(0, ep_existing.clone().try_into().unwrap()),
         Blake3::hash(&[0u8]),
         2 * ep_existing,
@@ -316,7 +324,10 @@ pub async fn test_get_child_at_epoch_at_root() -> Result<(), HistoryTreeNodeErro
     }
     let latest_ep = root.get_latest_epoch();
     assert!(latest_ep.unwrap_or(0) == 4, "Latest epochs don't match!");
-    assert!(root.epochs.len() == 3, "Ask yourself some pressing questions, such as: Why are there random extra epochs in the root?");
+    assert!(
+        root.birth_epoch < root.last_epoch,
+        "How is the last epoch not higher than the birth epoch after an udpate?"
+    );
 
     Ok(())
 }
@@ -327,11 +338,23 @@ pub async fn test_get_child_at_epoch_at_root() -> Result<(), HistoryTreeNodeErro
 async fn test_insert_single_leaf_root() -> Result<(), HistoryTreeNodeError> {
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(0u64)).await?;
-    let new_leaf =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b0u64, 1u32), 1, &[0u8], 0, 0).await?;
+    let new_leaf = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b0u64, 1u32),
+        &[0u8],
+        NodeLabel::root(),
+        0,
+    )
+    .await?;
 
-    let leaf_1 =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b1u64, 1u32), 2, &[1u8], 0, 0).await?;
+    let leaf_1 = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b1u64, 1u32),
+        &[1u8],
+        NodeLabel::root(),
+        0,
+    )
+    .await?;
     root.write_to_storage(&db).await?;
 
     let mut num_nodes = 1;
@@ -371,15 +394,32 @@ async fn test_insert_single_leaf_root() -> Result<(), HistoryTreeNodeError> {
 async fn test_insert_single_leaf_below_root() -> Result<(), HistoryTreeNodeError> {
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(0u64)).await?;
-    let new_leaf =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b00u64, 2u32), 1, &[0u8], 0, 1).await?;
+    let new_leaf = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b00u64, 2u32),
+        &[0u8],
+        NodeLabel::root(),
+        1,
+    )
+    .await?;
 
-    let leaf_1 =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b11u64, 2u32), 2, &[1u8], 0, 2).await?;
+    let leaf_1 = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b11u64, 2u32),
+        &[1u8],
+        NodeLabel::root(),
+        2,
+    )
+    .await?;
 
-    let leaf_2 =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b10u64, 2u32), 3, &[1u8, 1u8], 0, 3)
-            .await?;
+    let leaf_2 = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b10u64, 2u32),
+        &[1u8, 1u8],
+        NodeLabel::root(),
+        3,
+    )
+    .await?;
 
     let leaf_0_hash = Blake3::merge(&[
         Blake3::merge(&[Blake3::hash(&[]), Blake3::hash(&[0b0u8])]),
@@ -439,19 +479,41 @@ async fn test_insert_single_leaf_below_root() -> Result<(), HistoryTreeNodeError
 async fn test_insert_single_leaf_below_root_both_sides() -> Result<(), HistoryTreeNodeError> {
     let db = InMemoryDb::new();
     let mut root = get_empty_root::<Blake3, _>(&db, Option::Some(0u64)).await?;
-    let new_leaf =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b000u64, 3u32), 1, &[0u8], 0, 0).await?;
+    let new_leaf = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b000u64, 3u32),
+        &[0u8],
+        NodeLabel::root(),
+        0,
+    )
+    .await?;
 
-    let leaf_1 =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b111u64, 3u32), 2, &[1u8], 0, 0).await?;
+    let leaf_1 = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b111u64, 3u32),
+        &[1u8],
+        NodeLabel::root(),
+        0,
+    )
+    .await?;
 
-    let leaf_2 =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b100u64, 3u32), 3, &[1u8, 1u8], 0, 0)
-            .await?;
+    let leaf_2 = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b100u64, 3u32),
+        &[1u8, 1u8],
+        NodeLabel::root(),
+        0,
+    )
+    .await?;
 
-    let leaf_3 =
-        get_leaf_node::<Blake3, _>(&db, NodeLabel::new(0b010u64, 3u32), 4, &[0u8, 1u8], 0, 0)
-            .await?;
+    let leaf_3 = get_leaf_node::<Blake3, _>(
+        &db,
+        NodeLabel::new(0b010u64, 3u32),
+        &[0u8, 1u8],
+        NodeLabel::root(),
+        0,
+    )
+    .await?;
 
     let leaf_0_hash = Blake3::merge(&[
         Blake3::merge(&[Blake3::hash(&[]), Blake3::hash(&[0b0u8])]),
@@ -531,9 +593,8 @@ async fn test_insert_single_leaf_full_tree() -> Result<(), HistoryTreeNodeError>
         let new_leaf = get_leaf_node::<Blake3, _>(
             &db,
             NodeLabel::new(i.clone(), 3u32),
-            leaves.len() as u64,
             &i.to_ne_bytes(),
-            0,
+            NodeLabel::root(),
             7 - i,
         )
         .await?;
